import 'dart:async';

import 'package:meta/meta.dart';
import 'package:rxdart/rxdart.dart';
import 'functions.dart';

abstract class Event {
  const Event();
}

class EventBus {
  EventBus()
      : _events = PublishSubject(),
        _handlers = {};
  final PublishSubject<Event> _events;
  final Map<Type, List<_Handler>> _handlers;

  @visibleForTesting
  Observable<Event> get eventsStream => _events;

  void emit<TEvent extends Event>(TEvent event) {
    if (!_events.isClosed) {
      _events.add(event);
    }
  }

  void addListener<TEvent extends Event>(EventHandler<TEvent> handler) {
    final handlers = _getOrCreateHandlersOf<TEvent>();

    final wrapper = (Event e) {
      return handler(e);
    };

    // ignore: cancel_subscriptions
    final _sub = _events.where((e) => e is TEvent).listen(wrapper);
    handlers.add(_Handler(handler, _sub));
  }

  void addAsyncListener<TEvent extends Event>(
      AsyncEventHandler<TEvent> handler) {
    final handlers = _getOrCreateHandlersOf<TEvent>();

    // ignore: cancel_subscriptions
    final _sub = _events.where((e) => e is TEvent).listen((e) {
      //ignore: unawaited_futures
      handler(e);
    });

    handlers.add(_Handler(handler, _sub));
  }

  List<_Handler<dynamic>> _getOrCreateHandlersOf<TEvent>() {
    final type = _typeOf<TEvent>();

    var handlers = _handlers[type];
    if (handlers == null) {
      _handlers[type] = [];
      handlers = _handlers[type];
    }

    return handlers;
  }

  void removeListener<TEvent extends Event>(
      DynamicEventHandler<TEvent> handler) {
    final handlers = _getOrCreateHandlersOf<TEvent>();

    final h = handlers.firstWhere(
      (t) => t.handler == handler,
      orElse: () => null,
    );
    if (h != null) {
      h.subscription.cancel();
      handlers.remove(h);
    }
  }

  @mustCallSuper
  void dispose() {
    _events.close();

    for (var list in _handlers.values) {
      for (var handler in list) {
        handler.subscription.cancel();
      }
    }
  }
}

class _Handler<T> {
  const _Handler(this.handler, this.subscription);
  final dynamic handler;
  final StreamSubscription<T> subscription;
}

Type _typeOf<T>() => T;
