import 'package:flutter_test/flutter_test.dart';
import 'package:frebu/frebu.dart';
import 'package:mockito/mockito.dart';

class FakeOnDataEffect extends Mock implements OnDataEffect<int, int> {}

class FakeOnExceptionEffect extends Mock
    implements OnExceptionEffect<int, int> {}

void main() {
  group('Test onException method', () {
    test('onException gets correct arguments', () async {
      final onException = FakeOnExceptionEffect();

      final action = AsyncAction<int, int>((int a) async => throw a)
          .onException(onException)
          .mapException((a, e, [t]) => a);

      final result = await action.run(5);

      verify(onException.onException(5, 5, any));

      expect(result, 5);
    });

    test('onException does not call when exceptions are not occured', () async {
      final onException = FakeOnExceptionEffect();

      final action = AsyncAction<int, int>((int a) async => a)
          .onException(onException)
          .mapException((arg, e, [t]) => arg);

      final result = await action.run(5);

      verifyNever(onException.onException(any, any, any));
      verifyNever(onException.onException(any, any));

      expect(result, 5);
    });
  });

  group('Test onData method', () {
    test('onData gets correct result', () async {
      final onData = FakeOnDataEffect();

      final action = AsyncAction((int a) async => a + a).onData(onData);

      final result = await action.run(5);

      verify(onData.onData(5, 10));

      expect(result, 10);
    });

    test('onData does not call when there is any exception', () async {
      final onData = FakeOnDataEffect();

      final action = AsyncAction<int, int>((int a) async => throw 'a')
          .onData(onData)
          .mapException((arg, e, [t]) => arg);

      final result = await action.run(5);

      verifyNever(onData.onData(any, any));

      expect(result, 5);
    });
  });

  group('Test mapException method', () {
    test(
      'mapException catches exception and transform them to any other object',
      () async {
        final action = AsyncAction<int, int>((int a) async => throw '2')
            .mapException((arg, e, [t]) => int.parse(e));

        final result = await action.run(5);

        expect(result, 2);
      },
    );

    test('mapException gets correct args', () async {
      final action = AsyncAction<int, int>((int a) async => throw a)
          .mapException((arg, e, [t]) => arg);

      final result = await action.run(5);

      expect(result, 5);
    });
  });

  group('Test map method', () {
    test('map changes returned type from int to string', () async {
      final action = AsyncAction((int a) => Future.value(a))
          .map((args, res) => res.toString());

      final result = await action.run(5);

      expect(result, "5");
    });

    test('map gets correct args', () async {
      final action = AsyncAction((int arg) => Future.value(arg + arg))
          .map((args, res) => args);

      final result = await action.run(5);

      expect(result, 5);
    });

    test('map throws away inner exceptions', () async {
      final action = AsyncAction((int a) => Future.value(a))
          .map((args, res) async => throw 'exception');

      expect(() => action.run(5), throwsA(isA<String>()));
    });
  });

  group('Test run method', () {
    test('run returns a value', () async {
      final action = AsyncAction((int arg) async => arg.toString());

      final result = await action.run(5);

      expect(result, "5");
    });

    test('run throws away inner exceptions', () async {
      final action = AsyncAction((int arg) async => throw 'exception');

      expect(() => action.run(5), throwsA(isA<String>()));
    });
  });
}
